import datetime
import logging
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Union

import confuse
from sqlalchemy import create_engine, engine
from sqlalchemy import func as sa_func
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import scoped_session, sessionmaker, Session

from aardvark.exceptions import CombineException, DatabaseException
from aardvark.persistence import PersistencePlugin
from aardvark.persistence.sqlalchemy.models import AdvisorData, AWSIAMObject, Base

log = logging.getLogger("aardvark")
session_type = Union[scoped_session, Session]


class SQLAlchemyPersistence(PersistencePlugin):
    sa_engine: engine = None
    session_factory: sessionmaker = None

    def __init__(
        self, alternative_config: confuse.Configuration = None, initialize: bool = True
    ):
        super().__init__(alternative_config=alternative_config)
        if initialize:
            self.init_db()

    def init_db(self):
        self.sa_engine = create_engine(self.config["sqlalchemy"]["database_uri"].get())
        self.session_factory = sessionmaker(
            autocommit=False,
            autoflush=False,
            bind=self.sa_engine,
            expire_on_commit=False,
        )
        session = self._create_session()
        Base.query = session.query_property()

        Base.metadata.create_all(bind=self.sa_engine)

    def _create_session(self) -> scoped_session:
        return scoped_session(self.session_factory)

    def teardown_db(self):
        Base.metadata.drop_all(bind=self.sa_engine)

    def create_iam_object(
        self, arn: str, last_updated: datetime.datetime
    ) -> AWSIAMObject:
        with self.session_scope() as session:
            item = AWSIAMObject(arn=arn, lastUpdated=last_updated)
            session.add(item)
        return item

    @contextmanager
    def session_scope(self):
        """Provide a transactional scope around a series of operations."""
        session: session_type = self._create_session()
        try:
            yield session
            log.debug("committing SQLAlchemy DB session")
            session.commit()
        except Exception as e:
            log.debug("exception caught, rolling back session: %s", e)
            session.rollback()
            raise
        finally:
            log.debug("closing SQLAlchemy DB session")
            session.close()

    def _combine_results(self, access_advisor_data: Dict[str, Any]) -> Dict[str, Any]:
        access_advisor_data.pop("page")
        access_advisor_data.pop("count")
        access_advisor_data.pop("total")
        usage: Dict[str, Dict] = dict()
        for arn, services in access_advisor_data.items():
            for service in services:
                namespace = service.get("serviceNamespace")
                last_authenticated = service.get("lastAuthenticated")
                if namespace not in usage:
                    usage[namespace] = service
                else:
                    count_entities = (
                        usage[namespace]["totalAuthenticatedEntities"]
                        + service["totalAuthenticatedEntities"]
                    )
                    if last_authenticated > usage[namespace]["lastAuthenticated"]:
                        usage[namespace] = service
                    usage[namespace]["totalAuthenticatedEntities"] = count_entities

        for namespace, service in usage.items():
            last_authenticated = service["lastAuthenticated"]
            if isinstance(last_authenticated, int):
                dt_last_authenticated = datetime.datetime.fromtimestamp(
                    last_authenticated / 1e3
                )
            elif isinstance(last_authenticated, str):
                dt_last_authenticated = datetime.datetime.strptime(
                    last_authenticated, "%Y-%m-%d %H:%M:%S.%f"
                )
            else:
                dt_last_authenticated = last_authenticated

            dt_starting = datetime.datetime.utcnow() - datetime.timedelta(days=90)
            usage[namespace]["USED_LAST_90_DAYS"] = dt_last_authenticated > dt_starting

        return usage

    def store_role_data(self, access_advisor_data: Dict[str, Any]):
        with self.session_scope() as session:
            if not access_advisor_data:
                log.warning(
                    "Cannot persist Access Advisor Data as no data was collected."
                )
                return

            arn_cache = {}
            for arn, data in access_advisor_data.items():
                if arn in arn_cache:
                    item = arn_cache[arn]
                else:
                    item = self.get_or_create_iam_object(arn)
                    arn_cache[arn] = item
                for service in data:
                    self.create_or_update_advisor_data(
                        item.id,
                        service["LastAuthenticated"],
                        service["ServiceName"],
                        service["ServiceNamespace"],
                        service.get("LastAuthenticatedEntity"),
                        service["TotalAuthenticatedEntities"],
                        session=session,
                    )

    def get_role_data(
        self,
        page: int = 0,
        count: int = 0,
        combine: bool = False,
        phrase: str = "",
        arns: Optional[List[str]] = None,
        regex: str = "",
        session: session_type = None,
    ) -> Dict[str, Any]:
        offset = (page - 1) * count if page else 0
        limit = count
        session = session or self._create_session()
        # default unfiltered query
        query = session.query(AWSIAMObject)

        try:
            if phrase:
                query = query.filter(AWSIAMObject.arn.ilike("%" + phrase + "%"))

            if arns:
                query = query.filter(
                    sa_func.lower(AWSIAMObject.arn).in_([arn.lower() for arn in arns])
                )

            if regex:
                query = query.filter(AWSIAMObject.arn.regexp(regex))

            total = query.count()

            if offset:
                query = query.offset(offset)

            if limit:
                query = query.limit(limit)

            items = query.all()
        except Exception as e:
            raise DatabaseException("Could not retrieve roles from database: %s", e)

        if not items:
            items = session.query(AWSIAMObject).offset(offset).limit(limit).all()

        values = dict(page=page, total=total, count=len(items))
        for item in items:
            item_values = []
            for advisor_data in item.usage:
                item_values.append(
                    dict(
                        lastAuthenticated=advisor_data.lastAuthenticated,
                        serviceName=advisor_data.serviceName,
                        serviceNamespace=advisor_data.serviceNamespace,
                        lastAuthenticatedEntity=advisor_data.lastAuthenticatedEntity,
                        totalAuthenticatedEntities=advisor_data.totalAuthenticatedEntities,
                        lastUpdated=item.lastUpdated,
                    )
                )
            values[item.arn] = item_values

        if combine and total > len(items):
            raise CombineException(
                "Error: Please specify a count of at least {}.".format(total)
            )
        elif combine:
            return self._combine_results(values)

        return values

    def create_or_update_advisor_data(
        self,
        item_id: int,
        last_authenticated: int,
        service_name: str,
        service_namespace: str,
        last_authenticated_entity: str,
        total_authenticated_entities: int,
        session: session_type = None,
    ):
        session = session or self._create_session()
        service_name = service_name[:128]
        service_namespace = service_namespace[:64]
        item = None
        try:
            item = (
                session.query(AdvisorData)
                .filter(AdvisorData.item_id == item_id)
                .filter(AdvisorData.serviceNamespace == service_namespace)
                .scalar()
            )
        except SQLAlchemyError as e:
            log.error(
                f"Database error: {e} item_id: {item_id} serviceNamespace: {service_namespace}"
            )

        if not item:
            item = AdvisorData(
                item_id=item_id,
                lastAuthenticated=last_authenticated,
                serviceName=service_name,
                serviceNamespace=service_namespace,
                lastAuthenticatedEntity=last_authenticated_entity,
                totalAuthenticatedEntities=total_authenticated_entities,
            )
            try:
                session.add(item)
            except SQLAlchemyError as e:
                log.error(f"failed to add AdvisorData item to session: {e}")
                raise
            return

        # sqlite will return a string for item.lastAuthenticated, so we parse that into a datetime
        if isinstance(item.lastAuthenticated, str):
            ts = datetime.datetime.strptime(
                item.lastAuthenticated, "%Y-%m-%d %H:%M:%S.%f"
            )
        else:
            ts = item.lastAuthenticated

        if last_authenticated > ts:
            item.lastAuthenticated = last_authenticated
            try:
                session.add(item)
            except SQLAlchemyError as e:
                log.error(f"failed to add AdvisorData item to session: {e}")
                raise

        elif last_authenticated < ts:
            """
            lastAuthenticated is obtained by calling get_service_last_accessed_details() method of the boto3 iam client.
            When there is no AA data about a service, the lastAuthenticated key is missing from the returned dictionary.
            This is perfectly valid, either because the service in question was not accessed in the past 365 days or
            the entity granting  access to it was created recently enough that no AA data is available yet (it can take up to
            4 hours for this to happen).
            When this happens, the AccountToUpdate._get_job_results() method will set lastAuthenticated to 0.
            Usually we don't want to persist such an entity, with one exception: there's already a recorded, non-zero lastAuthenticated
            timestamp persisted for this item. That means the service was accessed at some point in time, but now more than 365 passed since
            the last access, so AA no longer returns a timestamp for it.
            """
            if last_authenticated == 0:
                log.warning(
                    "Previously seen object not accessed in the past 365 days "
                    "(got null lastAuthenticated from AA). Setting to 0. "
                    f"Object {item.item_id} service {item.serviceName} previous timestamp {item.lastAuthenticated}"
                )
                item.lastAuthenticated = 0
                try:
                    session.add(item)
                except SQLAlchemyError as e:
                    log.error(f"failed to add AdvisorData item to session: {e}")
                    raise
            else:
                log.error(
                    f"Received an older time than previously seen for object {item.item_id} service {item.serviceName} ({last_authenticated} < {item.lastAuthenticated})!"
                )

    def get_or_create_iam_object(self, arn: str):
        with self.session_scope() as session:
            try:
                item = (
                    session.query(AWSIAMObject).filter(AWSIAMObject.arn == arn).scalar()
                )
            except SQLAlchemyError as e:
                log.error(f"failed to retrieve IAM object: {e}")
                raise

            added = False
            if not item:
                item = AWSIAMObject(arn=arn, lastUpdated=datetime.datetime.utcnow())
                added = True
            else:
                item.lastUpdated = datetime.datetime.utcnow()

            try:
                session.add(item)
            except SQLAlchemyError as e:
                log.error(f"failed to add AWSIAMObject item to session: {e}")
                raise

            # we only need a refresh if the object was created
            if added:
                try:
                    session.commit()
                    session.refresh(item)
                except SQLAlchemyError as e:
                    log.error(f"failed to create IAM object: {e}")
                    raise
            return item



from typing import Any, Dict, List, Optional

import confuse

from aardvark.plugins import AardvarkPlugin


class PersistencePlugin(AardvarkPlugin):
    def __init__(self, alternative_config: confuse.Configuration = None):
        super().__init__(alternative_config=alternative_config)

    def init_db(self):
        raise NotImplementedError()

    def teardown_db(self):
        raise NotImplementedError()

    def get_role_data(
        self,
        page: int = 0,
        count: int = 0,
        combine: bool = False,
        phrase: str = "",
        arns: Optional[List[str]] = None,
        regex: str = "",
    ) -> Dict[str, Any]:
        raise NotImplementedError()

    def store_role_data(self, access_advisor_data: Dict[str, Any]) -> None:
        raise NotImplementedError()

    def _combine_results(self, access_advisor_data: Dict[str, Any]) -> Dict[str, Any]:
        raise NotImplementedError()



import asyncio
import datetime
import logging
from typing import Any, Dict, Union

from asgiref.sync import sync_to_async
import confuse
from cloudaux.aws.sts import boto3_cached_conn

from aardvark.exceptions import AccessAdvisorException
from aardvark.retrievers import RetrieverPlugin

log = logging.getLogger("aardvark")


class AccessAdvisorRetriever(RetrieverPlugin):
    def __init__(self, alternative_config: confuse.Configuration = None):
        super().__init__("access_advisor", alternative_config=alternative_config)

    async def _generate_service_last_accessed_details(self, iam_client, arn):
        """Call IAM API to create an Access Advisor job."""
        result = await sync_to_async(iam_client.generate_service_last_accessed_details)(
            Arn=arn
        )
        return result["JobId"]

    async def _get_service_last_accessed_details(self, iam_client, job_id):
        """Retrieve Access Advisor job results. Do an exponential backoff if the job is not complete."""
        attempts = 0
        while attempts < 10:
            details = await sync_to_async(iam_client.get_service_last_accessed_details)(
                JobId=job_id
            )
            job_status = details.get("JobStatus")
            if job_status == "COMPLETED":
                return details
            elif job_status == "IN_PROGRESS":
                # backoff sleep and try again
                await asyncio.sleep(2 ** attempts)
                continue
            else:
                error = details.get("Error") or "no error details provided"
                raise AccessAdvisorException(f"Access Advisor job failed: {error}")

    @staticmethod
    def _get_account_from_arn(arn: str) -> str:
        """Return the AWS account ID from an ARN."""
        return arn.split(":")[4]

    @staticmethod
    def _transform_result(
        service_last_accessed: Dict[str, Union[str, int, datetime.datetime]]
    ) -> Dict[str, Union[str, int]]:
        """'Transform' Access Advisor result, which really just means convert the datetime to a timestamp."""
        last_authenticated = service_last_accessed.get("LastAuthenticated")

        # Convert from datetime to timestamp
        if last_authenticated:
            last_authenticated = int(last_authenticated.timestamp() * 1000)
        else:
            last_authenticated = 0

        service_last_accessed["LastAuthenticated"] = last_authenticated
        return service_last_accessed

    async def run(self, arn: str, data: Dict[str, Any]) -> Dict[str, Any]:
        """Retrieve Access Advisor data for the given ARN and add the results to `data["access_advisor"]`."""
        log.debug(f"running {self} for {arn}")
        account = self._get_account_from_arn(arn)
        conn_details: Dict[str, str] = {
            "account_number": account,
            "assume_role": self.config["aws"]["rolename"].as_str(),
            "session_name": "aardvark",
            "region": self.config["aws"]["region"].as_str() or "us-east-1",
            "arn_partition": self.config["aws"]["arn_partition"].as_str() or "aws",
        }
        iam_client = await sync_to_async(boto3_cached_conn)("iam", **conn_details)
        try:
            job_id = await self._generate_service_last_accessed_details(iam_client, arn)
        except iam_client.exceptions.NoSuchEntityException as e:
            log.info(f"ARN {arn} no longer exists in AWS IAM")
            return data

        aa_details = await self._get_service_last_accessed_details(iam_client, job_id)
        result = map(self._transform_result, aa_details["ServicesLastAccessed"])
        data["access_advisor"] = list(result)
        return data



# ensure absolute import for python3
from __future__ import absolute_import

from flask import Blueprint, abort, jsonify, request

from aardvark.persistence.sqlalchemy import SQLAlchemyPersistence

advisor_bp = Blueprint("advisor", __name__)
session = SQLAlchemyPersistence()._create_session()


@advisor_bp.teardown_request
def shutdown_session(exception=None):
    session.remove()


# undocumented convenience pass-through so we can query directly from browser
@advisor_bp.route("/advisors")
def get():
    return post()


@advisor_bp.route("/advisors")
def post():
    """Get access advisor data for role(s)
    Returns access advisor information for role(s) that match filters
    ---
    consumes:
      - 'application/json'
    produces:
      - 'application/json'

    parameters:
      - name: page
        in: query
        type: integer
        description: return results from given page of total results
        required: false
      - name: count
        in: query
        type: integer
        description: specifies how many results should be return per page
        required: false
      - name: combine
        in: query
        type: boolean
        description: combine access advisor data for all results [Default False]
        required: false
      - name: phrase
        in: query
        type: string
        description: TODO
        required: false
      - name: regex
        in: query
        type: string
        description: TODO
        required: false
      - name: arn
        in: query
        type: string
        description: TODO
        required: false

    definitions:
      AdvisorData:
        type: object
        properties:
          lastAuthenticated:
            type: number
          lastAuthenticatedEntity:
            type: string
          lastUpdated:
            type: string
          serviceName:
            type: string
          serviceNamespace:
            type: string
          totalAuthenticatedEntities:
            type: number
      QueryBody:
        type: object
        properties:
          phrase:
            type: string
          regex:
            type: string
          arn:
            type: array
            items: string
      Results:
        type: array
        items:
          $ref: '#/definitions/AdvisorData'

    responses:
      200:
        description: Query successful, results in body
        schema:
          $ref: '#/definitions/AdvisorData'
      400:
        description: Bad request - error message in body
    """
    try:
        page = int(request.args.get("page", 1))
        count = int(request.args.get("count", 30))
        combine = request.args.get("combine", type=str, default="false")
        combine = combine.lower() == "true"
        phrase = request.args.get("phrase")
        regex = request.args.get("regex", default=None)
        arns = request.args.get("arn")
        arns = arns.split(",") if arns else []
    except Exception as e:
        raise abort(400, str(e))

    values = SQLAlchemyPersistence().get_role_data(
        page=page,
        count=count,
        combine=combine,
        phrase=phrase,
        arns=arns,
        regex=regex,
        session=session,
    )

    return jsonify(values)



import yaml

from aardvark.configuration import create_config


def test_create_config(temp_config_file):
    create_config(
        aardvark_role="role",
        swag_bucket="bucket",
        swag_filter="filter",
        swag_service_enabled_requirement="service",
        arn_partition="aws",
        sqlalchemy_database_uri="sqlite://////////////hi.db",
        sqlalchemy_track_modifications=True,
        num_threads=99,
        region="us-underground-5",
        filename=temp_config_file,
    )

    with open(temp_config_file, "r") as f:
        file_data = yaml.safe_load(f)

    assert file_data["aws"]["rolename"] == "role"
    assert file_data["aws"]["region"] == "us-underground-5"
    assert file_data["aws"]["arn_partition"] == "aws"
    assert file_data["swag"]["bucket"] == "bucket"
    assert file_data["swag"]["filter"] == "filter"
    assert file_data["swag"]["service_enabled_requirement"] == "service"
    assert file_data["updater"]["num_threads"] == 99
    assert file_data["sqlalchemy"]["database_uri"] == "sqlite://////////////hi.db"
    assert file_data["sqlalchemy"]["track_modifications"]



import configparser

import confuse

CONFIG: confuse.Configuration = confuse.Configuration("aardvark", __name__)


def create_config(
    aardvark_role: str = "",
    swag_bucket: str = "",
    swag_filter: str = "",
    swag_service_enabled_requirement: str = "",
    arn_partition: str = "",
    sqlalchemy_database_uri: str = "",
    sqlalchemy_track_modifications: bool = False,
    num_threads: int = 5,
    region: str = "",
    filename: str = "generated.yaml",
):
    if aardvark_role:
        CONFIG["aws"]["rolename"] = aardvark_role
    if arn_partition:
        CONFIG["aws"]["arn_partition"] = arn_partition
    if region:
        CONFIG["aws"]["region"] = region
    if swag_bucket:
        CONFIG["swag"]["bucket"] = swag_bucket
    if swag_filter:
        CONFIG["swag"]["filter"] = swag_filter
    if swag_service_enabled_requirement:
        CONFIG["swag"]["service_enabled_requirement"] = swag_service_enabled_requirement
    if sqlalchemy_database_uri:
        CONFIG["sqlalchemy"]["database_uri"] = sqlalchemy_database_uri
    if sqlalchemy_track_modifications:
        CONFIG["sqlalchemy"]["track_modifications"] = sqlalchemy_track_modifications
    if num_threads:
        CONFIG["updater"]["num_threads"] = num_threads
    with open(filename, "w") as f:
        f.write(CONFIG.dump(full=False))


def get_config(component: str) -> configparser.SectionProxy:
    """
    Returns a different config set based on the provided environment.
    This is used for testing.
    """
    return CONFIG[component]


def get_config_from_file(filepath: str, component: str) -> configparser.SectionProxy:
    """
    This returns an entirely different configuration value, and utilizes the environemnt provided
    as an index into that configuration file. This is used via the CLI.
    """
    alternative_config = configparser.ConfigParser()
    alternative_config.read(filepath)
    return alternative_config[component]


def open_config(filepath: str):
    CONFIG.read(filepath)



r"""
              ,-~~~~-,
        .-~~~;        ;~~~-.
       /    /          \    \
      {   .'{  O    O  }'.   }
       `~`  { .-~~~~-. }  `~`
            ;/        \;
           /'._  ()  _.'\
          /    `~~~~`    \
         ;                ;
         {                }
         {     }    {     }
         {     }    {     }
         /     \    /     \
        { { {   }~~{   } } }
         `~~~~~`    `~~~~~`
           (`"======="`)
           (_.=======._)

Good boy.
"""

import logging
from typing import Any, Dict

import confuse

from aardvark.plugins import AardvarkPlugin

log = logging.getLogger("aardvark")


class RetrieverPlugin(AardvarkPlugin):
    _name: str

    def __init__(self, name: str, alternative_config: confuse.Configuration = None):
        super().__init__(alternative_config=alternative_config)
        self._name = name

    async def run(self, arn: str, data: Dict[str, Any]) -> Dict[str, Any]:
        raise NotImplementedError()

    @property
    def name(self):
        return self._name

    def __str__(self):
        return f"Retriever({self.name})"



SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"
SQLALCHEMY_TRACK_MODIFICATIONS = False

# Use a set to store ARNs that are constantly failing.
# Aardvark will only log these errors at the INFO level
# instead of the ERROR level
FAILING_ARNS = set()
# FAILING_ARNS = {'ASDF', 'DEFG'}



"""
name: aardvark
description: Multi-Account AWS IAM Access Advisor API
author: Patrick Kelley, Travis McPeak, Patrick Sanders
maintainer: Patrick Sanders
contact: aardvark-maintainers@netflix.com
"""
from setuptools import setup

setup(
    name="aardvark",
    python_requires="~=3.8",
    versioning="dev",
    setup_requires="setupmeta",
    entry_points={"console_scripts": ["aardvark = aardvark.manage:main"]},
)



import confuse

from aardvark.configuration import CONFIG


class AardvarkPlugin:
    def __init__(self, alternative_config: confuse.Configuration = None):
        if alternative_config:
            self.config = alternative_config
        else:
            self.config = CONFIG
